#include "display.h"
#include <limits>

// ,      
//  .
Display display_device;

//     ,   .
//        
//     ,  
//    .
static Img img;

//    ;   :
//   - , 
//   LCD;
//  , ,  RAM,    ;
//     .
int display_init()
{
	display_device.init();
	display_device.hw_reset();
	display_device.display_on(1);
	display_device.clear_screen();
	display_device.display_sleep_out(1);

	display_device.color=COLOR_GREEN;
	display_device.print(0, 0, "****************");
	display_device.print(1, 0, "***  FC-001  ***");
	display_device.print(2, 0, "****************");
	display_device.print(18, 0, "----------------");
	display_device.print(19, 0, "http://rotr.info");
	
	display_device.color=COLOR_WHITE;
	display_device.print(F_ROW, 0, "F:");
	display_device.print(H_ROW, 0, "H:");
	display_device.print(D_ROW, 0, "D:");
	display_device.print(T_ROW, 0, "T:");
	display_device.print(I_ROW, 0, "I:");
	
	display_device.color=COLOR_YELLOW;

	return 0;
}

//    ,    
//      .
int ilog10(uint32_t a)
{
	if(a>=1000000000)
		return 10;
	uint32_t p=10;
	int r=1;
	while(true)
	{
		if(a<p)
			return r;
		r++;
		p*=10;
	}
}

//        .
template<typename T>
T iexp(T x, int n)
{
	if(x==0&&n==0)
		return std::numeric_limits<T>::quiet_NaN();
	if(n==0)
		return 1;
	if(x==0||x==1)
		return x;
	if(n<0)
	{
		n=-n;
		if(n<0)
			return 1/x*iexp(x, -(n+1));

		return 1/iexp(x, n);
	}

	T p=x;
	T r=1;
	unsigned int k=n;
	while(true)
	{
		if(k&1)
			r*=p;
		k>>=1;
		if(k==0)
			return r;
		p*=p;
	}
}

//    (a)     
//       nd  .
//      ,  
//  ,     ,   
// nd ,    , ..    
//   nd+1       nd+2 
// .
int ftos(
		 float a,			//  .
		 unsigned int nd,	//    
		 char *buf			//    (   nd+2).
		 )
{
	//    9  (    
	//  ).
	if(nd>9)
		nd=9;

	//      1  .
	int nd2=nd;
	if(nd2<9)
		nd2++;

	//   ,   ,  .
	int sgn=0;
	if(a<0)
	{
		sgn=-1;
		a=-a;
	}
	else if(a>0)
		sgn=1;

	const char *special=0;
	//if(a==std::numeric_limits<float>::infinity())
	if(isinff(a))
		special=(sgn>0)?("inf"):("-inf");
	else if(isnanf(a))
	{
		special="NaN";
	}

	if(special!=0)
	{
		if(nd<2)
			special="#";
		char *p=buf;
		while(*special!=0)
			*(p++)=*(special++);
		return p-buf;
	}

	//    ,     .
	uint32_t d=floorf(a);
	int n=ilog10(d);

	//        
	//        (nd).
	if(n>nd)
		return 0;

	// ,       , 
	//       .
	int n2=nd2-n;

	uint32_t k=iexp(10U, n2);

	a-=d;
	d=d*k+floorf(a*k);

	// .
	if(nd2>nd)
	{
		bool up=d%10>=5;
		d/=10;
		n2--;
		uint32_t d0=d;
		if(up)
			d++;
		if(ilog10(d0)!=ilog10(d))
		{
			n++;
			d/=10;
			n2--;
		}
		if(n>nd)
			return 0;
	}

	//    .
	char *p=buf;
	for(int i=0; i<nd; i++)
	{
		if(i==n2)
			*(p++)='.';
		*(p++)=d%10+'0';
		d/=10;
	}

	if(sgn<0)
		*(p++)='-';

	int r=p-buf;

	//        ,
	//     .
	while(buf<p)
	{
		char t=*buf;
		*(buf++)=*(--p);
		*p=t;
	}
	return r;
}

//     (   );
//   row,    col    .
void display_print_blank(int row, int col)
{
	if(img.empty())
		img.reset(128, 8);

	img.bg_color=display_device.bg_color;

	img.erase();

	display_device.x=col*8+X_START;
	display_device.y=row*8+Y_START;

	while(display_device.x<Y_STOP)
		display_device.img_out(img.get_buf(), img.get_width(), img.get_height());
}

//      (  
//  ).
// row, col -    ,    ;
// a -  ;
// nd -  ,      
//   ;
// xp - ,  3        
//  (..., -3 - , 0 -  , 3 - , ...);
// sfx -      
//  (0   ,     );
// force_sign -     true,   
//     (-),     (+).
void display_print(int row, int col, float a, uint32_t nd, int xp, char *sfx, bool force_sgn)
{
	if(img.empty())
		img.reset(128, 8);

	img.color=display_device.color;
	img.bg_color=display_device.bg_color;

	img.erase();

	const char tbl[]={'n', 'u', 'm', ' ', 'k', 'M', 'G'};
	char buf[12];
	bool is_spec_num=isinff(a)||isnanf(a);
	if(!is_spec_num)
		a*=iexp(10.0, -xp);
	if(is_spec_num)
		sfx=0;
	int n=ftos(a, nd, buf);
	if(n>0)
	{
		char cs=(a>0)?'+':' ';
		if(a>=0&&force_sgn)
			img.text_out(cs);
		img.text_out(buf, n);

		char cx='#';
		if(xp%3==0)
		{
			xp/=3;
			if(xp>=-3&&xp<=3)
				cx=tbl[xp+3];
		}

		if(cx!=' ')
		{
			img.text_out(' ');
			img.text_out(cx);
		}

		if(cx!='#'&&sfx!=0&&*sfx!=0)
		{
			if(xp==0)
				img.text_out(' ');

			img.text_out(sfx);
		}
	}

	display_device.x=col*8+X_START;
	display_device.y=row*8+Y_START;

	display_device.img_out(img.get_buf(), img.get_width(), img.get_height());
}

//      a  
// a=b*10**xp,
//  xp  3,
// 1<=b<1000.
int get_xp(float a)
{
	if(a==0||isnanf(a)||isinff(a))
		return 0;
	if(a<0)
		a=-a;
	if(a>=1000000000)
		return 9;
	if(a>=1000000)
		return 6;
	if(a>=1000)
		return 3;
	if(a>=1)
		return 0;
	if(a>=0.001)
		return -3;
	if(a>=0.000001)
		return -6;
	return -9;
}

//      a  
// a=b*10**xp,  xp  3.
//     ,   
//   b:
// 0.9<b<1100;
//      ,   
//  (a)  " "    (a_prev).
int get_xp(float a, float a_prev)
{
	if(a==0||isnanf(a)||isinff(a))
		return 0;
	int xp1=get_xp(a_prev);
	int xp2=get_xp(a);
	if(xp1==xp2)
		return xp1;
	if((a>0)-(a_prev>0)!=0)
		return xp2;
	a=fabsf(a);
	a_prev=fabsf(a_prev);
	float k=(a>a_prev)?a_prev/a:a/a_prev;
	if(k<0.9)
		return xp2;
	return xp1;
}
